import inspect
from ast import parse, NodeTransformer, copy_location, Attribute, Name, Load, Is, Store, Assign, Compare, If, \
    FunctionDef, arguments, fix_missing_locations, Param, Subscript, Index, Num, Str, Expr, Call
from functools import partial
from itertools import takewhile
from logging import getLogger
from os.path import commonprefix, dirname, join, abspath

from typing import Callable, Tuple, Any, Dict, List, Union

from roro_ioc.container import IOCContainer
from roro_ioc.container_field_registry import get_fast_retrieval_context, get_fast_retrieval_resource_handle

_logger = getLogger(__name__)


class SourceCodeInaccessibleError(ValueError):
    pass


def _get_source(callable_arg):
    source = inspect.getsource(callable_arg)

    start_indent = ''.join(takewhile(lambda l: l.isspace(), source))
    len_start_indent = len(start_indent)

    first_line_no = callable_arg.func_code.co_firstlineno

    source_stripped_prefix = ('\n' * (first_line_no - 1)) + \
                             ('\n'.join(line[len_start_indent:] for line in source.split('\n')))

    try:
        return parse(source_stripped_prefix)
    except SyntaxError:
        raise SourceCodeInaccessibleError('Could not parse source code for function {}. Retrieved {}'.format(
            callable_arg, source))


class _ManglePrivateMembers(NodeTransformer):
    def __init__(self, class_name):
        self.class_name = class_name

    # noinspection PyPep8Naming
    def visit_Attribute(self, node):
        # Make sure descendants of this node are visited too
        self.generic_visit(node)

        attribute_name = node.attr
        if self._should_mangle(attribute_name):
            mangled_name = self._get_mangled_name(attribute_name)
            return copy_location(
                Attribute(value=node.value, attr=mangled_name, ctx=node.ctx),
                node
            )
        else:
            return node

    def _should_mangle(self, attribute_name):
        return attribute_name.startswith('__') and (not attribute_name.endswith('__')) and self.class_name

    def _get_mangled_name(self, attribute_name):
        mangled_class_name = '_{}'.format(self.class_name.lstrip('_'))
        mangled_name = '{}{}'.format(mangled_class_name, attribute_name)
        return mangled_name


class _InjectParameters(NodeTransformer):
    def __init__(self, parameters, default_argument_name):
        self.parameters = parameters  # type: Tuple[Tuple[basestring, int], ...]
        self.injected_arguments_set = frozenset(parameter[0] for parameter in self.parameters)
        # Schema: (argument_name, resource_name, resource_handle)
        self.default_argument_name = default_argument_name  # type: basestring

    # noinspection PyPep8Naming
    def visit_FunctionDef(self, node):
        function_name = node.name

        INTERNAL_CONTEXT_NAME = '___INJECT_CONTEXT_INTERNAL'
        INTERNAL_RESOURCES_NAME = '___INJECT_CONTEXT_INTERNAL_RESOURCES'

        last_existing_arg = node.args.args[-1]  # non-empty because we inject somewhere
        new_args = node.args.args + [copy_location(Name(id=INTERNAL_CONTEXT_NAME, ctx=Param()), last_existing_arg)]

        last_existing_default = node.args.defaults[-1]
        new_defaults = node.args.defaults + [copy_location(Name(id=self.default_argument_name, ctx=Load()),
                                                           last_existing_default)]

        # Format: (Name(id='a', ctx=Param()), Name(id='foo', ctx=Load()))
        def generate_default_nodes():
            for (argument, default_value) in zip(node.args.args[-len(node.args.defaults):], node.args.defaults):
                if argument.id in self.injected_arguments_set:
                    yield (argument.id, default_value)

        default_nodes_mapping = dict(generate_default_nodes())

        def _generate_assignment((arg_name, arg_resource_handle)):
            # type: (Tuple[basestring, int]) -> If
            """
            We have a function that looks like:
            def do_something(param, model_=INJECTED):
                <...>

            We insert into its beginning a statement like
                ___INJECT_CONTEXT_INTERNAL_RESOURCES = ___INJECT_CONTEXT_INTERNAL.resources
                if model_ is INJECTED:
                    model_ = ___INJECT_CONTEXT_INTERNAL_RESOURCES[3]
                    if model is ___INJECT_CONTEXT_INTERNAL:    # means that no resource is available
                        ___INJECT_CONTEXT_INTERNAL_RESOURCES.flag_missing('model_')

            Code outside of this function sets a global variable _INJECTED__model to point at the right thing.
            """
            target_attribute = Subscript(value=Name(id=INTERNAL_RESOURCES_NAME, ctx=Load()),
                                         slice=Index(value=Num(n=arg_resource_handle)), ctx=Load())

            consequence = [
                Assign(targets=[Name(id=arg_name, ctx=Store())],
                       value=target_attribute),
                If(test=Compare(left=Name(id=arg_name, ctx=Load()), ops=[Is()],
                                comparators=[Name(id=INTERNAL_CONTEXT_NAME, ctx=Load())]),
                   body=[Expr(value=Call(func=Attribute(value=Name(id=INTERNAL_CONTEXT_NAME, ctx=Load()),
                                                        attr='flag_missing', ctx=Load()),
                                         keywords=[],
                                         starargs=None,
                                         kwargs=None,
                                         args=[Str(s=arg_name)]))],
                   orelse=[])
            ]  # type: List[Union[Assign, If]

            return If(test=Compare(left=Name(id=arg_name, ctx=Load()), ops=[Is()],
                                   comparators=[default_nodes_mapping[arg_name]]),
                      body=consequence,
                      orelse=[])

        first_body_element = node.body[0]
        align_with_first_body_element = partial(copy_location, old_node=first_body_element)

        new_body = [align_with_first_body_element(
            Assign(targets=[Name(id=INTERNAL_RESOURCES_NAME, ctx=Store())],
                   value=Attribute(value=Name(id=INTERNAL_CONTEXT_NAME,
                                              ctx=Load()), attr='resources', ctx=Load())))]

        new_body.extend(map(align_with_first_body_element, map(_generate_assignment, self.parameters)))
        new_body.extend(node.body)

        result = copy_location(
            FunctionDef(name=function_name,
                        args=arguments(
                            args=new_args,
                            vararg=node.args.vararg,
                            kwarg=node.args.kwarg,
                            defaults=new_defaults),
                        body=new_body,
                        decorator_list=[]),  # Note that no decorators are applied, as @inject has to be the first
            node
        )
        return result


def rewrite_ast(type_or_callable, class_name, injectable_arguments_tuple, arg_to_ioc_container):
    # type: (Callable, basestring, Tuple[Tuple[basestring, basestring, Any], ...], Dict[basestring, IOCContainer])
    globals_dict = type_or_callable.func_globals
    ast_structure = _get_source(type_or_callable)
    function_name = ast_structure.body[0].name

    injected_arguments = tuple((argument_name,
                                get_fast_retrieval_resource_handle(arg_to_ioc_container[resource_name], resource_name))
                               for (argument_name, resource_name, _) in injectable_arguments_tuple)
    default_value_name = '__INJECT___FAST_RETRIEVAL_CONTEXT'
    _InjectParameters(injected_arguments, default_value_name).visit(ast_structure)
    _ManglePrivateMembers(class_name).visit(ast_structure)
    fix_missing_locations(ast_structure)

    locals_dict = {default_value_name: get_fast_retrieval_context()}
    eval(compile(ast_structure, filename=inspect.getfile(type_or_callable), mode="exec"),
         globals_dict, locals_dict)

    return locals_dict.get(function_name)


def is_suitable_for_rewriting(type_or_callable):
    try:
        target_path = abspath(inspect.getsourcefile(type_or_callable))
        project_root = abspath(join(dirname(__file__), '../../../'))
        common = commonprefix((project_root, target_path))
        return common == project_root
    except:
        _logger.exception('Exception checking function %s', type_or_callable)
        return False



from abc import ABCMeta, abstractproperty

from typing import FrozenSet


class CannotBeProvided(ValueError):
    pass


class IOCContainer(object):
    __metaclass__ = ABCMeta

    @abstractproperty
    def provides(self):
        # type: ()->FrozenSet[basestring]
        pass

    @abstractproperty
    def provided(self):
        # type: ()->object
        pass



import inspect
import itertools
import threading
from contextlib import contextmanager

import attr
from attr.exceptions import NotAnAttrsClassError
from attr.validators import instance_of
from cached_property import cached_property
from typing import FrozenSet

from roro_ioc.container import IOCContainer
from roro_ioc.container_field_registry import get_fast_retrieval_context, register_ioc_container, \
    get_fast_retrieval_resource_handle
from roro_ioc.exceptions import InvalidPayload, CannotArmTwice


def _validate_condition(o, a, v):
    return attr.fields(v)


_STRUCTURED_LOCAL = threading.local()


@attr.attrs(hash=False)
class InstanceIOCContainer(IOCContainer):
    injected_resource_type = attr.attrib(validator=_validate_condition)  # type: type
    allow_idempotent_arming = attr.attrib(validator=instance_of(bool))  # type: bool

    @cached_property
    def provides(self):
        # type: () -> FrozenSet[basestring]
        try:
            iterable_attrs = (field.name for field in attr.fields(self.injected_resource_type))
        except NotAnAttrsClassError:
            iterable_attrs = ()

        def condition(p):
            return inspect.ismethoddescriptor(p) or inspect.ismemberdescriptor(p) or inspect.isdatadescriptor(p)

        iterable_method_descriptors = (name for (name, dontcare) in
                                       inspect.getmembers(self.injected_resource_type, condition)
                                       if not name.startswith('_'))

        result = frozenset(itertools.chain(iterable_attrs, iterable_method_descriptors))

        return result

    def _validate_payload(self, payload):
        if not isinstance(payload, self.injected_resource_type):
            raise InvalidPayload()
        return True

    @contextmanager
    def arm(self, payload):
        self._validate_payload(payload)

        fast_retrieval_context = get_fast_retrieval_context()
        existing = _STRUCTURED_LOCAL.__dict__.get(self)
        if existing is not None:
            if (self.allow_idempotent_arming and
                    existing is payload):
                yield
                # And do nothing else
            else:
                raise CannotArmTwice()
        else:  # Is currently empty
            _integrate_resources(self, fast_retrieval_context, payload)
            _STRUCTURED_LOCAL.__dict__[self] = payload
            try:
                yield

            finally:
                del _STRUCTURED_LOCAL.__dict__[self]
                _cleanup_resources(self, fast_retrieval_context)

    @property
    def provided(self):
        return _STRUCTURED_LOCAL.__dict__.get(self)


# fast_retrieval_context is used as a placeholder for resources that are not currently provided
def _integrate_resources(ioc_container, fast_retrieval_context, payload):
    if not hasattr(fast_retrieval_context, 'resources'):
        fast_retrieval_context.resources = []
    for resource_name in ioc_container.provides:
        handle = get_fast_retrieval_resource_handle(ioc_container, resource_name)
        current_length = len(fast_retrieval_context.resources)
        if handle >= current_length:
            fast_retrieval_context.resources.extend([fast_retrieval_context] * (1 + handle - current_length))
        assert fast_retrieval_context.resources[handle] is fast_retrieval_context
        fast_retrieval_context.resources[handle] = getattr(payload, resource_name)


# fast_retrieval_context is used as a placeholder for resources that are not currently provided
def _cleanup_resources(ioc_container, fast_retrieval_context):
    for resource_name in ioc_container.provides:
        handle = get_fast_retrieval_resource_handle(ioc_container, resource_name)
        fast_retrieval_context.resources[handle] = fast_retrieval_context


def create_ioc_container(injected_resource_type, allow_idempotent_arming=False):
    # type: (type, bool)->InstanceIOCContainer
    result = InstanceIOCContainer(injected_resource_type,
                                  allow_idempotent_arming)
    register_ioc_container(result)
    return result



import inspect
import new
from functools import wraps
from logging import getLogger
from os import environ

from typing import Optional

from roro_ioc.ast_injection import rewrite_ast, is_suitable_for_rewriting
from roro_ioc.container import IOCContainer
from roro_ioc.exceptions import NoSourceForArgument, NoDefaultValueForArgument, DoubleProvidingProhibited
from roro_ioc.exceptions import NoValuesProvided
from roro_ioc.factory_inspection import extract_factory_specification, FactorySpecification
from roro_ioc.injected_tag import INJECTED

_USE_WRAPPING_INJECTOR = environ.get('TWG_WRAPPING_INJECTOR')

_logger = getLogger()


def __get_class_name():
    stack_frame = inspect.currentframe().f_back.f_back
    self = stack_frame.f_locals.get('self', None)
    if stack_frame.f_code.co_name == '<module>' or '__module__' not in stack_frame.f_locals:
        # only classes have this:
        return None
    else:
        return stack_frame.f_code.co_name


def inject(*injectors):
    return __inject_internal('', injectors, __get_class_name())


def inject_(*injectors):
    return __inject_internal('_', injectors, __get_class_name())


def inject_methods(*injectors):
    return __inject_methods_internal('', injectors)


def inject_methods_(*injectors):
    return __inject_methods_internal('_', injectors)


def __member_condition(x):
    return inspect.ismethod(x) or inspect.isfunction(x)


def __unbind(method):
    # TODO: python2 specific code
    # See: http://stackoverflow.com/questions/14574641/python-get-unbound-class-method
    if not inspect.ismethod(method) or method.im_self is None:
        return method
    return new.instancemethod(method.im_func, None, method.im_class)


def __inject_methods_internal(suffix, injectors):
    assert all(isinstance(injector, IOCContainer) for injector in injectors)

    def treat(subject_class, function_object):
        unbound = __unbind(function_object)
        result = __inject_internal(suffix, injectors, subject_class.__name__)(unbound)
        if result == unbound:  # but always different than the original for bound methods
            return function_object
        else:
            return result

    def decorate(subject_class):
        patch_list = tuple((name,
                            inspect.isfunction(method),
                            inspect.ismethod(method) and method.__self__ is subject_class,
                            method,
                            treat(subject_class, method))
                           for (name, method) in inspect.getmembers(subject_class, __member_condition))
        for (name, is_static, is_class, original_value, value) in patch_list:
            # If we hadn't decorated the method, it does not need re-decoration
            if original_value == value:
                continue
            if is_static:
                value = staticmethod(value)
            elif is_class:
                value = classmethod(value)
            setattr(subject_class, name, value)

        return subject_class

    return decorate


def __inject_internal(suffix, injectors, class_name):
    suffix_length = len(suffix) if suffix else 0

    arg_to_ioc_container = {name: injector
                            for injector in injectors
                            for name in injector.provides}

    # Reverse order - first injector overrides
    already_seen = set()
    for injector in injectors:
        for name in injector.provides:
            if name in already_seen:
                raise DoubleProvidingProhibited('Resource %s provided twice, injectors %s',
                                                name, arg_to_ioc_container)
            already_seen.add(name)

    def correspondence(argument_name):
        # type: (basestring) -> Optional[basestring]
        # Matches 'mydata_' to 'mydata'
        if suffix:
            if not argument_name.endswith(suffix):
                return None  # is not marked for replacement
            result = argument_name[:-suffix_length]
        else:
            result = argument_name
        if result in arg_to_ioc_container:
            return result
        else:
            return None  # is not provided

    def decorate(type_or_callable):
        factory_specification = extract_factory_specification(type_or_callable)  # type: FactorySpecification

        injectable_arguments = {argument: corresponding
                                for (argument, corresponding) in
                                ((argument, correspondence(argument)) for argument in
                                 factory_specification.argument_names)
                                if corresponding is not None}

        # This is a mapping to e.g. 'mydata_' to 'mydata'

        def validate_treatment(argument_name):
            # type: (basestring)->bool
            _NULL = []
            result = factory_specification.argument_default_values.get(argument_name, _NULL)
            if result is _NULL:
                if argument_name in injectable_arguments:
                    raise NoDefaultValueForArgument(
                        'Argument {} in callable {} implied to be injectable, but no default value was specified'.
                            format(argument_name, type_or_callable))
            else:
                if result is INJECTED and argument_name not in injectable_arguments:
                    raise NoSourceForArgument('Cannot inject argument {} into callable {}, injectable arguments {}'.
                                              format(argument_name, type_or_callable,
                                                     injectable_arguments.keys()))

        for element in factory_specification.argument_names:
            validate_treatment(element)

        # We do this only after calculating injection_treatment, so we can catch variables marked as INJECT who do not
        # have a source
        if len(injectable_arguments) == 0:
            return type_or_callable  # Nothing to do here

        arg_to_position = {arg_name: index
                           for (index, arg_name) in enumerate(factory_specification.argument_names)
                           if arg_name in injectable_arguments}

        # optimize for retrieval
        injectable_arguments_tuple = tuple((argument, corresponding, arg_to_position[argument])
                                           for (argument, corresponding) in injectable_arguments.iteritems())

        def _raise_missing(argument, corresponding):
            if argument not in factory_specification.argument_default_values or \
                    factory_specification.argument_default_values[argument] is INJECTED:
                raise NoValuesProvided('Cannot provide for value {}'.format(corresponding))

        if not _USE_WRAPPING_INJECTOR:
            if inspect.isfunction(type_or_callable) or \
                    inspect.ismethod(type_or_callable) or inspect.ismethoddescriptor(type_or_callable):
                if is_suitable_for_rewriting(type_or_callable):
                    return rewrite_ast(type_or_callable,
                                       class_name,
                                       injectable_arguments_tuple,
                                       arg_to_ioc_container)
                else:
                    _logger.warn('Tried to decorate %s, but it is unsuitable. Proceeding with fallback',
                                 type_or_callable)

        @wraps(type_or_callable)
        def substitute_parameters(*args, **kwargs):
            for (argument, corresponding, position_for_argument) in injectable_arguments_tuple:
                if position_for_argument < len(args) or argument in kwargs:
                    continue  # do not override this variable
                provided = arg_to_ioc_container[corresponding].provided
                try:
                    kwargs[argument] = getattr(provided, corresponding)
                except AttributeError:  # provided is None
                    # If it has a default, and no provider is available for the data, let the default be used implicitly
                    _raise_missing(argument, corresponding)

            return type_or_callable(*args, **kwargs)

        return substitute_parameters

    return decorate



from roro_ioc.container import IOCContainer
from roro_ioc.exceptions import NoSourceForArgument, NoValuesProvided
from roro_ioc.inject import (inject,
                             inject_,
                             inject_methods,
                             inject_methods_)
from roro_ioc.injected_tag import INJECTED, INJECTED_IF_AVAILABLE
from roro_ioc.instance_ioc_container import create_ioc_container



from inspect import getargspec
from logging import getLogger

import attr
from typing import Callable, Dict, Any, Tuple
from werkzeug.datastructures import ImmutableDict

_logger = getLogger(__name__)


@attr.attrs
class FactorySpecification(object):
    constructing_function = attr.attrib()  # type: Callable
    argument_names = attr.attrib(validator=attr.validators.instance_of(tuple))  # type: Tuple[basestring, ...]
    argument_default_values = attr.attrib(validator=attr.validators.instance_of(
        ImmutableDict))  # type: Dict[basestring, Any]


def _format_defaults(arg_names, defaults):
    defaults = defaults or ()
    defaults_matching = (attr.NOTHING,) * (len(arg_names) - len(defaults)) + defaults
    return ImmutableDict((arg_name, default_value) for (arg_name, default_value)
                         in zip(arg_names, defaults_matching)
                         if default_value != attr.NOTHING)


def extract_factory_specification_for_attrs(type_, allow_defaults):
    fields = [field for field in attr.fields(type_)
              if allow_defaults or field.default == attr.NOTHING]

    argument_names = tuple(field.name for field in fields)
    defaults = ImmutableDict((field.name, field.default) for field in fields
                             if field.default != attr.NOTHING)
    subject_callable = type_
    return FactorySpecification(subject_callable, argument_names, defaults)


def _extract_factory_specification_for_functions(functional_object):
    argument_names = ()
    defaults = ImmutableDict()
    try:
        argspec = getargspec(functional_object)
        argument_names = tuple(argspec.args)
        defaults = _format_defaults(argument_names, argspec.defaults)
    except TypeError:
        _logger.exception('Could not get argument specs for %s', functional_object)

    return FactorySpecification(functional_object, argument_names, defaults)


def extract_factory_specification(type_or_factory, allow_defaults=True):
    # type: (Callable, bool) -> FactorySpecification
    if isinstance(type_or_factory, type):
        try:
            return extract_factory_specification_for_attrs(type_or_factory, allow_defaults)
        except attr.exceptions.NotAnAttrsClassError:
            subject_callable = type_or_factory.__init__
            # May still fail for built-in objects/slots/...
    else:
        subject_callable = type_or_factory

    return _extract_factory_specification_for_functions(subject_callable)



class NoSourceForArgument(TypeError):
    pass


class NoValuesProvided(ValueError):
    pass


class NoDefaultValueForArgument(ValueError):
    pass


class DoubleProvidingProhibited(TypeError):
    pass


class InvalidPayload(ValueError):
    pass


class CannotArmTwice(Exception):
    pass



from setuptools import setup

setup(
    name='roro_ioc',
    packages=['roro_ioc'],
    version='0.1.2',
    description='IOC Injection for python',
    author='Twiggle',
    author_email='oren@twiggle.com',
    url='https://github.com/twgOren/roroioc',
    keywords=['ioc'],
    classifiers=[],
    install_requires=[
        'attrs',
        'typing',
        'cached_property',
        'werkzeug',
    ]
)



import threading
from abc import ABCMeta

import attr
from typing import Dict, Any

from roro_ioc.container import IOCContainer
from roro_ioc.exceptions import NoValuesProvided


@attr.attrs
class _ContainerFieldRegistry(object):
    _mapping = attr.attrib(
        validator=attr.validators.instance_of(dict),
        default=attr.Factory(dict))  # type: Dict[Tuple[IOCContainer, basestring], int]

    # TODO: handle the leakage of IOC Containers Contexts (not their contents - those should not leak!)
    def add(self, ioc_container):
        for resource_name in ioc_container.provides:
            self._mapping[(ioc_container, resource_name)] = len(self._mapping)

    def get(self, ioc_container, field):
        return self._mapping[(ioc_container, field)]


_IOC_CONTAINER_FIELD_REGISTRY = _ContainerFieldRegistry()


def get_fast_retrieval_resource_handle(ioc_container, resource_name):
    return _IOC_CONTAINER_FIELD_REGISTRY.get(ioc_container, resource_name)


class ResourcesHolder(object):
    __metaclass__ = ABCMeta

    resources = {}  # type: Dict[basestring, Any]


_CONTAINER_FIELDS = threading.local()  # type: ResourcesHolder


def _flag_missing(field_name):
    raise NoValuesProvided('Mandatory field <{}> was not provided'.format(field_name))


_CONTAINER_FIELDS.flag_missing = _flag_missing


def register_ioc_container(ioc_container):
    _IOC_CONTAINER_FIELD_REGISTRY.add(ioc_container)


def get_fast_retrieval_context():
    # type: ()->ResourcesHolder
    return _CONTAINER_FIELDS



import attr
from decorator import contextmanager
from roro_ioc.caching import static_memoize_weak_result
from typing import Any

from roro_ioc.instance_ioc_container import create_ioc_container, InstanceIOCContainer


@static_memoize_weak_result
def create_direct_injector(injected_type, injected_instance):
    # type: (type, Any)->DirectInjector
    assert isinstance(injected_instance, injected_type)
    return DirectInjector(create_ioc_container(injected_type),
                          injected_instance)


@attr.attrs
class DirectInjector(object):
    __structured_injector = attr.attrib(validator=attr.validators.instance_of(InstanceIOCContainer))
    __injector_configuration = attr.attrib()

    @contextmanager
    def arm(self):
        with self.__structured_injector.arm(self.__injector_configuration):
            yield

    def inject(self, to_decorate):
        def decorated(*args, **kwargs):
            with self.__structured_injector.arm(self.__injector_configuration):
                return self.__structured_injector.inject(to_decorate)(*args, **kwargs)

        return decorated
